library fields.default_editors;

import 'dart:html';
import 'package:tags/tags.dart';
import 'field.dart';
import 'field_editor.dart';
import 'field_transform.dart';
import 'theme.dart';

void registerDefaultFieldEditors() {  
  if (!FieldEditor.isDefaultRegistered)
    FieldEditor.registerDefault(DefaultEditor);  
  if (!FieldEditor.isRegistered(bool)) 
    FieldEditor.register(bool, BoolEditor);
  if (!FieldEditor.isRegistered(DateTime))
    FieldEditor.register(DateTime, DateTimeEditor);
  if (!FieldEditor.isRegistered(num))
    FieldEditor.register(num, NumberEditor);
}

class UseRange extends FieldAnnotation {
  final num min, max, step;
  const UseRange({this.min, this.max, this.step});
  void applyTo(InputElement input) {
    if (min != null) input.min = min.toString();      
    if (max != null) input.max = max.toString();      
    if (step != null) input.step = step.toString();
  }
}

class UseValues extends FieldAnnotation {
  final List values;
  const UseValues(this.values);
}

class BoolEditor extends FieldEditor<bool> {
  CheckboxInputElement _input;
  Element get element => _input;
  
  BoolEditor(Field f) : super(f) {
    _input = input(() {
      style()
      ..scoped = true
      ..text = resources.style; 
    })
    ..type = 'checkbox'
    ..onChange.listen((_) {
      this.value = _input.checked;
    });
    
    render();
  }
      
  void render() {
    _input
    ..disabled = isValueReadOnly
    ..checked = value;
  }
}

class ColorEditorResources extends FieldEditorResources {
  String get style => super.style + 
"""
.picker {  
  float:left;
}
.valuebox { 
  overflow:hidden; 
}
""";
  const ColorEditorResources();
}

class ColorEditor extends FieldEditor<String> {
  InputElement _picker;
  InputElement _display;
  Element _element;
  Element get element => _element;
  
  ColorEditor(Field f) : super(f, const ColorEditorResources()) {
    _element = span(() {
      style()
      ..scoped = true
      ..text = resources.style;   
      _picker = input()
          ..type = 'color'
          ..classes = ['picker']
          ..onChange.listen((_) {
            _display.value = _picker.value;         
            this.value = _picker.value;
          });
      div(() {
        _display = input()
            ..type = 'text'
            ..disabled = true;
      })..classes=['valuebox'];
    });
    
    render();
  }
  
  void render() {
    _picker
    ..disabled = isValueReadOnly
    ..value = value;
    _display.value = value;
  }
}

class DateTimeEditor extends FieldEditor<DateTime> {
  DateInputElement _input;
  Element get element => _input;
  
  DateTimeEditor(Field f) : super(f) {
    _input = input(() {
      style()
      ..scoped = true
      ..text = resources.style; 
    })
    ..type = 'date' // TODO(rms): investigate 'datetime-local'
    ..onChange.listen((_) {
      this.value = _input.valueAsDate;
    });
    
    render();
  }
  
  void render() {
    _input
    ..disabled = isValueReadOnly
    ..valueAsDate = value;
  }
}

class DefaultEditor extends FieldEditor<String> {
  InputElement _input;
  Element get element => _input;
    
  DefaultEditor(Field f) : super(f) {
    _input = input(() {
      style()
      ..scoped = true
      ..text = resources.style; 
    })
    ..type = 'text'
    ..onInput.listen((_) {
      try {
        this.value = _input.value;
        _input.classes.remove('error');
      } catch(_) {
        _input.classes.add('error');
      }
    });
    
    render();
  }
  
  void render() {
    _input
    ..disabled = isValueReadOnly
    ..value = value;
  }
}

class NumberEditor extends FieldEditor<num> {
  InputElement _input;
  Element get element => _input;
  
  NumberEditor(Field f) : super(f) {
    _input = input(() {
      style()
      ..scoped = true
      ..text = resources.style; 
    })
    ..type = 'number'
    ..onChange.listen((_) {
      this.value = _input.valueAsNumber;
    });
    
    final match = (a) => a is UseRange;
    if (f.annotations.any(match)) {
      UseRange range = f.annotations.firstWhere(match);
      range.applyTo(_input);
    }
    
    render();
  }
  
  void render() {
    _input
    ..disabled = isValueReadOnly
    ..valueAsNumber = value;
  }
}

class RangeEditorResources extends FieldEditorResources {
  String get style => super.style + 
"""
.range {  
  float:left;
}
.valuebox { 
  overflow:hidden; 
}
""";
  const RangeEditorResources();
}

class RangeEditor extends FieldEditor<num> {
  InputElement _range;
  InputElement _display;
  Element _element;
  Element get element => _element;
  
  RangeEditor(Field f) : super(f, const RangeEditorResources()) {
    _element = span(() {
      style()
      ..scoped = true
      ..text = resources.style;   
      
      _range = input()
          ..type = 'range'
          ..classes = ['range']
          ..onChange.listen((_) {
            _display.value = _range.value; 
            this.value = _range.valueAsNumber;
          });
      
      div(() {
        _display = input()
            ..type = 'text'
            ..disabled = true;
      })
      ..classes=['valuebox'];
    });

    final match = (a) => a is UseRange;
    if (f.annotations.any(match)) {
      UseRange range = f.annotations.firstWhere(match);
      range.applyTo(_range);
    }
    
    render();
  }
  
  void render() {
    _range
    ..disabled = isValueReadOnly
    ..valueAsNumber = value;
    _display.value = value.toString();
  }
}

class SelectEditorResources extends FieldEditorResources {
  String get style => 
"""
select {
  background-color:transparent;
  border: 0;
  border-radius: 0;
  color:${fieldsTheme.colors.text};
}
option {
  background: ${fieldsTheme.colors.fieldBackground};
}
""";
  const SelectEditorResources();
}

class SelectEditor extends FieldEditor<String> {
  final List _optionValues = new List();
  SelectElement _select;
  Element get element => _select;
      
  SelectEditor(Field f) : super(f, const SelectEditorResources()) {
    _select = select(() {
      style()
      ..scoped = true
      ..text = resources.style; 
      
      final isUseValues = (a) => a is UseValues;
      if (f.annotations.any(isUseValues)) {
        UseValues useValues = f.annotations.firstWhere(isUseValues);
        _optionValues.addAll(useValues.values);        
        for (var o in _optionValues) {
          // TODO(rms): "".runtimeType != String
          // https://code.google.com/p/dart/issues/detail?id=7949
          // I think what we actually want to do is to have `transformField`
          // check `isAssignableTo` rather than direct equality
          String valueString = transformField(o, /*o.runtimeType*/String, String);
          option()
          ..text = valueString
          ..value = valueString;
        }
      }
    })
    ..onChange.listen((_) {
      this.value = _select.value;
    });
    
    render();
  }
  
  void render() {
    _select
    ..disabled = isValueReadOnly
    ..value = value;
  }
}
