library fields.field;

import 'dart:async';
import 'dart:mirrors';
import 'field_list.dart';
import 'field_transform.dart';
import 'mirrors.dart';

/// An annotation that may be applied to an instance variable or getter of an 
/// object's class declaration.
/// 
/// A default set of field annotations are defined in this library.  This
/// annotation may be extended or implemented to provide metadata to custom
/// field editors.  Custom field annotations are made available to editors as
/// [Field.annotations].
abstract class FieldAnnotation {
  const FieldAnnotation();  
}

/// An annotation that declares a field should be ignored by editors.
///
/// A call to [new Field.forVariable] or [new Field.forGetter] will return 
/// `null` if this annotation is found in the field's declaration.
const FieldAnnotation IgnoreField = const _IgnoreField();
class _IgnoreField extends FieldAnnotation {
  const _IgnoreField();
}

/// An annotation that declares the display name for a field.
///
/// By default the display name of a field is the name of the variable or getter
/// as given by [Field.name].
class DisplayName extends FieldAnnotation {
  
  /// The display name for the annotated field.
  final String name;
  
  /// Create a constant display name annotation with the given [name].
  const DisplayName(this.name);
}

/// An annotation that declares a field to be read-only.
/// 
/// The default value of [Field.isReadOnly] is `true` for final variables or a
/// getter with no matching setter.  This annotation provides an override to 
/// declare that a field be considered read-only if the default value is 
/// `false`.
const FieldAnnotation ReadOnly = const _ReadOnly();
class _ReadOnly extends FieldAnnotation {
  const _ReadOnly();
}

/// An annotation that declares the use of a specific [FieldEditor] for a field.
/// 
/// By default a field editor is chosen based on a field's `runtimeType`.  This
/// annotation provides an override to declare that a field use an instance of
/// the specified [editorType].
class UseFieldEditor extends FieldAnnotation {
  
  /// The editor type to use for the annotated field.
  final Type editorType;
  
  /// Create a constant field editor annotation with the given [editorType].
  /// 
  /// The given [editorType] must extend or implement [FieldEditor].
  const UseFieldEditor(this.editorType);
}

/// An annotation that declares the use of a specific [FieldTransform] for a 
/// field.
/// 
/// By default a field transform is chosen based on a field's `runtimeType`.  
/// This annotation provides an override to declare that a field use the given
/// [transform] instead.
class UseFieldTransform extends FieldAnnotation {
  
  /// The transform to use for the annotated field.
  final FieldTransform transform;  
  
  /// Create a constant field transform annotation with the given [transform].
  const UseFieldTransform(this.transform);
}

/// A reflected instance variable or getter of an [Object].
class Field {
  
  static Map _parseMetadata(DeclarationMirror declaration) {
    assert(declaration != null);
    final metadata = new Map()
        ..[UseFieldEditor] = null
        ..[UseFieldTransform] = null
        ..[ReadOnly] = false
        ..[IgnoreField] = false
        ..[DisplayName] = null
        ..[FieldAnnotation] = <FieldAnnotation>[];
    declaration.metadata.forEach((InstanceMirror mirror) {
      final instance = mirror.reflectee;
      if (instance is UseFieldEditor) {
        // Disallow > 1 @UseFieldEditor
        assert(metadata[UseFieldEditor] == null);
        // TODO(rms): validate that the editor type isAssignableTo FieldEditor.
        metadata[UseFieldEditor] = reflectClass(instance.editorType);
      } else if (instance is UseFieldTransform) {
        // Disallow > 1 @UseFieldTransform
        assert(metadata[UseFieldTransform] == null);
        metadata[UseFieldTransform] = instance.transform;
      } else if (instance == ReadOnly) {
        metadata[ReadOnly] = true;
      } else if (instance == IgnoreField) {
        metadata[IgnoreField] = true;
      } else if (instance is DisplayName) {
        // Disallow > 1 @DisplayName
        assert(metadata[DisplayName] == null);
        metadata[DisplayName] = instance.name;
      } else if (instance is FieldAnnotation) {
        // Editor-specific annotations, store them in the field as-is.
        metadata[FieldAnnotation].add(instance);
      }
    });
    return metadata;
  }
  
  /// Creates a new field for the getter with the given [name] in the given 
  /// [owner] object.
  /// 
  /// An optional [getter] mirror and [parent] field may also be given.  In
  /// general these need not be specified; They are exposed as optional 
  /// parameters to allow editor implementations to optimize by caching and 
  /// reusing these values for multiple fields.
  factory Field.forGetter(
      InstanceMirror owner, 
      Symbol name, 
     {MethodMirror getter,
      Field parent}) {
    if (getter == null) {
      getter = owner.type.getters[name];
      if (getter == null) {
        throw new ArgumentError(
            '$name is not a getter of ${owner.type.qualifiedName}');
      }
    }
    final setter = new Symbol('${MirrorSystem.getName(name)}=');
    final metadata = _parseMetadata(getter);      
    if (metadata[IgnoreField]) {
      return null;
    }
    return new Field._(
        owner, 
        name, 
        typeOf(getter.returnType), 
        parent: parent,
        editor: metadata[UseFieldEditor],
        transform: metadata[UseFieldTransform],
        annotations: metadata[FieldAnnotation],
        displayName: metadata[DisplayName],
        isPrivate: getter.isPrivate,
        isReadOnly: !owner.type.setters.containsKey(setter) 
          || metadata[ReadOnly],
        isVariable: false);
  }
  
  /// Creates a new field for the instance variable with the given [name] in 
  /// the given [owner] object.
  /// 
  /// An optional [variable] mirror and [parent] field may also be given.  In
  /// general these need not be specified; They are exposed as optional 
  /// parameters to allow editor implementations to optimize by caching and 
  /// reusing these values for multiple fields.
  factory Field.forVariable(
      InstanceMirror owner, 
      Symbol name, 
     {VariableMirror variable,
      Field parent}) {
    if (variable == null) {
      variable = owner.type.variables[name];
      if (variable == null || variable.isStatic) {
        throw new ArgumentError(
            '$name is not an instance variable of ${owner.type.qualifiedName}');
      }
    }
    final metadata = _parseMetadata(variable);      
    if (metadata[IgnoreField]) {
      return null;
    }
    return new Field._(
        owner, 
        name, 
        typeOf(variable.type), 
        parent: parent,
        editor: metadata[UseFieldEditor],
        transform: metadata[UseFieldTransform],
        annotations: metadata[FieldAnnotation],
        displayName: metadata[DisplayName],
        isPrivate: variable.isPrivate,
        isReadOnly: variable.isFinal || metadata[ReadOnly]);
  }
  
  /// The name of this field as declared in the [owner]'s class declaration.
  final Symbol name;
  
  /// The display name of this field.
  /// 
  /// By default this is equal to the [name] of this field.  It may be overriden
  /// by annotating the field with a [DisplayName] annotation.
  final String displayName;
  
  /// The type of this field's [value].
  final Type type;
  
  /// Returns `true` if this field is declared as a getter.
  bool get isGetter => !isVariable;
  
  /// Returns `true` if this field is declared as private.
  final bool isPrivate;
  
  /// Returns `true` if this field is read-only.
  /// 
  /// A field is read-only if it reflects a `final` variable, a getter with no
  /// matching setter, or if it is annotated with [ReadOnly].
  final bool isReadOnly;
  
  /// Returns `true` if this field is declared as an instance variable.
  final bool isVariable;
  
  /// The parent of this field or `null` if this field has no parent.
  final Field parent;
  
  /// A mirror on the class of this field's editor; this is `null` by default.
  /// 
  /// By default, this is `null` and the choice of a field editor for this field
  /// is typically based on the [type].  If the field is annotated with 
  /// [UseFieldEditor] then this will be the reflected class of the annotated
  /// editor type.
  final ClassMirror editor;
  
  /// This field's preferred transform; this is `null` by default.
  /// 
  /// By default, this is `null` and the choice of a field transform for this
  /// field is chosen by an editor based on the [type].  If the field is
  /// annotated with [UseFieldTransform] then this will be the annotated 
  /// transform.
  final FieldTransform transform;
  
  /// Any field annotations that have been applied to this field outside of the
  /// set of field annotations defined in this library.
  /// 
  /// The set of field annotations defined in this library ([DisplayName], 
  /// [ReadOnly], etc...) have a direct effect on one or more properties of this
  /// field and are therefore not included here.
  /// 
  /// This collection provides a mechanism for annotating fields with custom
  /// metadata that may be accessed later by custom editors.
  final Iterable<FieldAnnotation> annotations;
  
  FieldList _children;
  /// The children of this field.
  /// 
  /// A field's children does not include those whose [type] is determined to be 
  /// `simple`.  The definition of `simple` used here is more inclusive than the 
  /// term as used in `dart:mirrors` and includes `Map`, `List`, `dynamic`, 
  /// `Type`, and `Object`.
  FieldList get children => _children;
  
  /// A mirror on the object instance that 'owns' this field.
  InstanceMirror owner;  
  
  Object _value;
  /// The current value of this field.
  Object 
    get value => _value;
    set value(Object value) {
      if(isReadOnly) {
        throw new StateError('Field $displayName is read-only.');
      }
      if (value != _value) {
        owner.setField(name, value);
        _value = value;
        _refreshChildren().then((_) {
          _values.add(value);
          if (parent != null) {
            parent._values.add(parent._value);
          }
        });
      }
    }
  
  final _values = new StreamController<Object>.broadcast();
  /// An observable stream of this field's values as they change.
  Stream<Object> get values => _values.stream;
  
  Field._(
      this.owner, 
      Symbol name, 
      this.type, 
     {this.parent, 
      this.editor, 
      this.transform,
      this.annotations: const [], 
      String displayName, 
      this.isPrivate: false, 
      this.isReadOnly: false,
      this.isVariable: true})
      : this.name = name
      , this.displayName = 
          (displayName == null) ? MirrorSystem.getName(name) : displayName {
    _children = new FieldList(this);
    _value = owner.getField(name).reflectee;
    if (!isSimpleType(type)) {
      _children.refreshSource(_value);
    }
  }
  
  /// Refresh this field's [value].
  /// 
  /// A field can only know about changes to its [value] that happen through
  /// this interface.  If another object changes the underlying value in some 
  /// other way then this method should be called.
  void refreshValue() {
    final value = owner.getField(name).reflectee;
    if (value != _value) {
      _value = value;
      _refreshChildren().then((_) => _values.add(value));
    }
  }
        
  Future _refreshChildren() {    
    if (children.isEmpty) {
      return new Future.value();
    }
    children.refreshSource(_value);
    return Future.forEach(children, (f) => new Future(f.refreshValue));
  }
}
